# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import time
from typing import Callable
from typing import Optional

from stopit import ThreadingTimeout

from mdbh import environ


def _create_lockfile(file: Path):
    file.with_suffix('.lock').touch()
    return


def _remove_lockfile(file: Path):
    file.with_suffix('.lock').unlink()
    return


def _lockfile_exists(file: Path) -> bool:
    """Check if .lock file exists for current file.
    The suffix of the file is replaced with .lock.

    Returns:
        True if .lock files exist, else False

    """
    return file.with_suffix(".lock").exists()


def _wait_for_lockfile(file: Path, timeout=120):
    """Waits, as long as a .lock file exists for the specified file.
    If no .lock file exists, returns directly.

    Args:
        file:    File to wait for.
        timeout: Number of seconds after which to timeout.

    Raises:
        TimeoutError after timeout seconds.

    """
    if _lockfile_exists(file):
        # Create a context manager for timeout control
        with ThreadingTimeout(seconds=timeout, swallow_exc=True) as mgr:
            assert mgr.state == mgr.EXECUTING
            # Wait, as long as .lock file exists
            while _lockfile_exists(file):
                print(f"Waiting for .lock file {file}.")
                time.sleep(1.0)

        if mgr.state == mgr.TIMED_OUT:
            raise TimeoutError(f"Timeout occurred while waiting for lockfile {file}.")
    return


def _lookup(filename: str) -> Optional[Path]:
    """Check whether file exists in cache folder.
    Files are automatically prefixed with "MDBH_".

    Args:
        filename: Name of file to lookup.

    Returns:
        If present, Path to cached file, else None.
    """
    file_path = Path(environ.get_cache_dir()) / filename
    file_path = file_path.with_name(f"MDBH_{file_path.name}")

    _wait_for_lockfile(file_path)

    if Path.is_file(file_path):
        print(f"Found in cache: {filename}")
        return file_path

    return None


def _save_call(filename: str,
               get_data: Callable[[], bytes],
               force: bool = False) -> Path:
    """Save file to cache from get_data callable.
    Files are automatically prefixed with "MDBH_".

    Args:
        filename: Name of the file to cache.

        get_data: Callable that retrieves binary data.

        force: Whether to force saving even if file exists.

    Raises:
        IOError when file already exists and force=False.
    """
    file_path = Path(environ.get_cache_dir()) / filename
    file_path = file_path.with_name(f"MDBH_{file_path.name}")

    if file_path.is_file() and not force:
        raise IOError(f"Cache file already exists: {file_path}")

    _create_lockfile(file_path)
    try:
        with open(file_path, "wb") as file:
            file.write(get_data())

    except KeyboardInterrupt:
        print("Aborting download. Removing temporary files.")
        file_path.unlink(missing_ok=True)

    except Exception as e:
        print("An exception occured. Removing temporary files.")
        file_path.unlink(missing_ok=True)
        raise e

    except:
        print("An unknown exception occured. Removing temporary files.")
        file_path.unlink(missing_ok=True)

    finally:
        # Always release lockfile
        _remove_lockfile(file_path)

    return file_path


def _save(filename: str,
          data: bytes,
          force: bool = False) -> Path:
    """Save file to cache.
    Files are automatically prefixed with "MDBH_".

    Args:
        filename: Name of the file to cache.

        data: Data bytes to cache.

        force: Whether to force saving even if file exists.

    Raises:
        IOError when file already exists and force=False.
    """
    get_data = lambda: data

    return _save_call(filename, get_data, force)


def get(filename: str,
        get_data: Callable[[], bytes],
        force: bool = False) -> Path:
    """Get filename from cache.
    If it does not yet exist in cache, use get_data to retrieve the data and
    cache the content.

    Args:
        filename: Filename to get.
        get_data: Callable to retrieve file content if file is not yet cached.
        force: Whether to force calling get_data and replacing the possibly cached file.

    Returns:
        Path to cached file.
    """
    cachefile = _lookup(filename)

    if cachefile is None or force:
        cachefile = _save_call(filename, get_data, force=force)

    return cachefile


def empty_cache():
    """Delete all cached files
    """
    cache_dir = Path(environ.get_cache_dir())
    cached_files = list(cache_dir.glob("MDBH_*"))
    print(f"Removing cached files {cached_files}.")
    [f.unlink() for f in cached_files]
    return



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

try:
    from mdbh._version import __version__
except ImportError:
    __version__ = "local"

__author__ = "Maximilian Schambach"

from . import environ
from .core import get_conf_databases
from .core import get_mongodb
from .core import get_uri
from .sacred import get_artifact
from .sacred import get_artifact_names
from .sacred import get_dataframe
from .sacred import get_metric
from .sacred import get_run_names
from .sacred import resolve_artifacts

environ.set_defaults()



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Optional, List

import numpy as np
import matplotlib.pyplot as plt

from pymongo.database import Database

from mdbh import get_metric


def plot_metric(db: Database,
                ids: List[int],
                label_y,
                label_x: Optional[str] = None,
                plot_norm: Optional[str] = None,
                avg: Optional[str] = None,
                avg_arg: Optional[float] = None,
                id_avg: bool = False,
                y_min: float = None,
                y_max: float = None,
                plt_show: bool = True,
                plt_kwargs: dict = {}):
    """Plot metrics of one or multiple run IDs.

    Args:
        db: Database instance.

        ids: List of run IDs to plot.

        label_y: Metric label of y-axis.

        label_x: Optional metric label of x-axis. If None, epoch number is used.

        plot_norm: Optional plotting norm.
                   Either "semilogx", "semilogy", or "loglog".
                   If None, regular plot is used.

        avg: Averaging of data.
             Either "MA" (moving average), "Gauss" (Gauss filtering) or
             "exp" (exponential weighting).

        avg_arg: Optional average parameter. For moving average, specifies
                   filter length, for Gauss filter, specifies sigma,
                   for exponential weighting, specifies exponential factor.

        id_avg: Whether to also average over the multiple run IDs.

        plt_show: Whether to show the plot.

    Returns:
        Matplotlib figure instance.
    """
    ma = None
    sigma = None
    beta = None
    if avg is not None:
        if avg.lower() == "ma":
            ma = avg_arg if avg_arg is not None else 5
        elif avg.lower() == "gauss":
            sigma = avg_arg if avg_arg is not None else 3
        elif avg.lower() == "exp":
            beta = avg_arg if avg_arg is not None else 3

    mpl_p = plt.semilogx if plot_norm == "semilogx" \
        else plt.semilogy if plot_norm == "semilogy" \
        else plt.loglog if plot_norm == "loglog" \
        else plt.plot

    y_dicts = [get_metric(db, id, label_y) for id in ids]
    y = [y_dict['values'] for y_dict in y_dicts]
    y_name = [y_dict['name'] for y_dict in y_dicts]

    if label_x is not None:
        x_dicts = [get_metric(db, id, label_x) for id in ids]
        x = [x_dict['values'] for x_dict in x_dicts]
        x_name = [x_dict['name'] for x_dict in x_dicts]
    else:
        x = [y_dict['steps'] for y_dict in y_dicts]
        x_name = ["Steps" for _ in y_dicts]

    lim_x = slice(0, -1)
    lim_y = lim_x
    if ma is not None and ma != 0:
        y = [np.convolve(tmp, (1 / ma) * np.ones(ma), 'same') for tmp in y]
        lim_x = slice(ma, -ma)
        lim_y = lim_x

    elif sigma is not None:
        from scipy.ndimage import gaussian_filter1d
        y = [gaussian_filter1d(tmp, sigma) for tmp in y]

    elif beta is not None:
        # Exponential weighting of values
        if len(ids) > 1:
            raise ValueError("Exponentially weighted smoothing only available "
                             "for single ID metrics.")

        avg = np.zeros(len(y))
        avg[0] = y[0]
        for i in range(1, len(avg)):
            avg[i] = beta * avg[i - 1] + (1 - beta) * y[i]

        norm = np.asarray([1 - beta**(i + 1) for i in range(len(avg))])
        y = (avg / norm).tolist()

    # Calculate avarage/std across multiple IDs
    y_std = None
    if id_avg and len(ids) > 1:
        # Check that all runs have the same x length
        x_lens = [len(i) for i in x]
        if len(set(x_lens)) != 1:
            raise ValueError("Cannot average runs with different lenghths of x axis.")
        y = np.asarray(y)
        y_std = np.std(y, axis=0, ddof=1)
        y = np.mean(y, axis=0)
        x = x[0]

    fig = plt.figure()
    if y_std is not None:
        mpl_p(x[lim_x], y[lim_y], label="ID average", **plt_kwargs)
        plt.fill_between(x[lim_x], (y - y_std)[lim_y], (y + y_std)[lim_y], alpha=0.2)
    else:
        for i, id in enumerate(ids):
            mpl_p(x[i][lim_x], y[i][lim_y], label=id, **plt_kwargs)
    plt.title(f"Metric plot for ID {ids}")
    plt.xlabel(x_name)
    plt.ylabel(y_name[0])

    if y_min is not None and y_max is not None:
        plt.ylim([y_min, y_max])

    if len(ids) > 1:
        plt.legend()

    if plt_show:
        plt.show()

    return fig



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import configparser
from pathlib import Path
from typing import List
from typing import Tuple
from typing import Union

from pymongo import MongoClient
from pymongo.database import Database


def get_uri(conf_path: Union[Path, str],
            db_name: str) -> str:
    """Creates MongoDB connection URI from configuration file
    and database name.

    Args:
        conf_path: Path to the .mongo.conf configuration file.

        db_name: Name of the database to connect to.

    Returns:
        uri, MongoDB connection URI.
    """
    # Get path
    p = Path(conf_path).expanduser().absolute()

    if not p.exists():
        raise ValueError(f"Path {p} doest not exist.")

    # Load MongoDB configuration
    config = configparser.ConfigParser()
    config.read(p)

    server = config['SERVER']['ip']
    port = config['SERVER']['port']

    if 'username' in config['USER'].keys() and 'password' in config['USER'].keys():
        user = config['USER']['username']
        pw = config['USER']['password']

        if "auth" in config['USER'].keys():
            auth = config['USER']['auth']
        else:
            auth = "SCRAM-SHA-1"

        if "auth_db" in config['USER'].keys():
            auth_db = config['USER']['auth_db']
        else:
            auth_db = "admin"

        uri = f"mongodb://{user}:{pw}@{server}:{port}/{db_name}?authSource={auth_db}&authMechanism={auth}"

    else:
        print("No username and password specified.")
        uri = f"mongodb://{server}:{port}/{db_name}"

    return uri


def get_conf_databases(conf_path: Union[Path, str]) -> List[str]:
    """Get the name of databases specified in .mongo.conf config file.
    If none specified, uses "sacred" as default.

    Args:
        conf_path: Path to the .mongo.conf configuration file.

    Returns:
        List of Sacred databases.

    """
    # Get path
    p = Path(conf_path).expanduser().absolute()

    if not p.exists():
        raise ValueError(f"Path {p} doest not exist.")

    # Load MongoDB configuration
    config = configparser.ConfigParser()
    config.read(p)

    if 'DATABASES' not in config.keys():
        return ["sacred"]

    return [db for db in config['DATABASES'].values()]


def get_client(conf_path: Union[Path, str],
               db_name: str = "sacred") -> MongoClient:
    """Get a MongoClient using the specified configuration."""

    uri = get_uri(conf_path, db_name)
    return MongoClient(uri)


def get_mongodb(conf_path: Union[Path, str],
                db_name: str = "sacred") -> Database:
    """Get a MongoDB Database instance.

    Args:
        conf_path: Path to MongoDB configuration file.

        db_name: Name of the database.

    Returns:
        A PyMongo Database instance.

    """
    client = get_client(conf_path, db_name)
    db = client.get_database()

    if not ('metrics' in db.collection_names() and 'runs' in db.collection_names()):
        print("WARNING: No 'metrics' or 'runs' collection found in database. "
              "Are you sure this is a Sacred database? "
              "If this is intended, you can safely ignore this message.")

    return db


def _wrap_ids(ids: Union[int, List[int], Tuple[int, ...]]) -> List[int]:
    """Wrap input ids into list of ints.

    Args:
        ids: Single or multiple ids.

    Returns:
        List of ids.
    """
    return list(ids) if isinstance(ids, (tuple, list, set)) else [ids]



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from pathlib import Path

import configparser

# System-independent path for MDBH config file
CONFIG_PATH = Path("~/.config/mdbh/mdbh.conf").expanduser()


def get():
    """Get the MDBH environment variables and values."""
    return {k: val for k, val in os.environ.items() if k[:5] == "MDBH_"}


def create_config_file():
    """Create default configuration file."""
    from tempfile import gettempdir

    print(f"No MDBH config file found at {CONFIG_PATH}. Creating default.")
    CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    config = configparser.ConfigParser()
    config['ENVIRON'] = dict(cache_dir=gettempdir(),
                             data_collection="binarydata")

    with open(CONFIG_PATH, 'w') as configfile:
        config.write(configfile)


def set_defaults():
    """Set the default MDBH environment variables and values."""

    if not CONFIG_PATH.exists():
        create_config_file()

    config = configparser.ConfigParser()
    config.read(CONFIG_PATH)

    # Read config file
    cache_dir = Path(config.get("ENVIRON", "cache_dir")).expanduser()
    set_cache_dir(str(cache_dir))

    data_collection = config.get("ENVIRON", "data_collection")
    set_data_collection(data_collection)


def get_cache_dir() -> str:
    """get the MDBH cache directory environment variable."""
    return os.environ['MDBH_CACHE_DIR']


def set_cache_dir(cache_dir: str):
    """Set the MDBH cache directory environment variable."""
    os.environ['MDBH_CACHE_DIR'] = cache_dir
    # Create directory and parents, if non-existent
    Path(cache_dir).mkdir(parents=True, exist_ok=True)


def get_data_collection() -> str:
    """get the MDBH data collection environment variable."""
    return os.environ['MDBH_DATA_COLLECTION']


def set_data_collection(data_collection: str):
    """get the MDBH data collection environment variable."""
    os.environ['MDBH_DATA_COLLECTION'] = data_collection



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from pathlib import Path
import pickle
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pandas as pd
from pymongo.database import Collection
from pymongo.database import Database

from mdbh import caching
from mdbh import data
from mdbh.core import _wrap_ids
from mdbh.data import _get_gridfs_data


def _get_dataframe_query(db: Database,
                         query: dict,
                         collection: Union[str, Collection]) -> pd.DataFrame:
    """Get a DataFrame from a MongoDB query.

    Args:
        db: Database instance to query.

        query: The query.

        collection: Collection to query.

    Returns:
        DataFrame of query result.
    """
    if type(db) != Database:
        raise TypeError("db needs to be a PyMongo Database instance. "
                        "Use mdbh.get_mongodb")
    # Query database and convert to DataFrame
    res = db[collection].find(query)
    df = pd.DataFrame(list(res))

    return df


def _get_run_ids(db: Database) -> List[int]:
    """Get all run IDs of a database.

    Args:
        db: Database instance to query.

    Returns:
        List of run IDs of the database.
    """
    return _get_dataframe_query(db, {}, "runs")['_id'].tolist()


def _get_dataframe(db: Database,
                   ids: Optional[Union[int, List[int], Tuple[int, ...]]] = None,
                   include_artifacts: bool = False) -> pd.DataFrame:
    """Get all experiments from a Sacred Database in a DataFrame.
    Metrics and configurations are concatenated into a single DataFrame for
    convenience.

    Args:
        db: Database instance to query.

        ids: Optional run ID or list of IDs to get.
             If None, all IDs are queried.

        include_artifacts: Whether to include a column for artifacts.
                           Artifacts can then be resolved via
                           :func:resolve_artifacts().

    Returns:
        DataFrame of the corresponding Database.
    """
    # Wrap ids into list
    ids = _wrap_ids(ids) if ids is not None else ids

    # Get all metric and run logs
    query = {} if ids is None else {"run_id": {"$in": ids}}
    df_metric = _get_dataframe_query(db, query, "metrics")

    query = {} if ids is None else {"_id": {"$in": ids}}
    df_runs = _get_dataframe_query(db, query, "runs")
    df_runs.index = df_runs['_id'].tolist()

    # Extract config dict into columns
    tmp = pd.DataFrame(df_runs['config'].tolist())
    tmp = tmp.rename(lambda x: f"config.{x}", axis=1)
    tmp.index = df_runs.index
    df_runs = pd.concat([df_runs, tmp], axis=1)

    # Extract experiment dict into columns
    tmp = pd.DataFrame(df_runs['experiment'].tolist())
    tmp = tmp.rename(lambda x: f"experiment.{x}", axis=1)
    tmp.index = df_runs.index
    df_runs = pd.concat([df_runs, tmp], axis=1)

    # Clean up DataFrame
    df_m_del_cols = ['_id', ]
    df_r_del_cols = ['experiment', 'format', 'command', 'start_time', 'meta',
                     'resources', 'info', 'heartbeat',
                     'result', 'stop_time', 'fail_trace', 'captured_out',
                     'omniboard']

    if not include_artifacts:
        df_r_del_cols += ['artifacts']

    for i in df_r_del_cols:
        if i in df_runs.columns.tolist():
            del df_runs[i]

    # When there are no recorded metrics, just return runs dataframe
    if len(df_metric) == 0:
        return df_runs.rename(columns={'_id': 'id'})

    # Set run_id (metrics), _id (runs) as index of DataFrames for later join
    df_metric.index = df_metric['run_id'].tolist()

    # Get all IDs
    id_list = df_runs['_id'].unique().tolist()

    # Get metric column names
    metrics = df_metric['name'].unique().tolist()
    cols = [f"metrics.{i}" for i in metrics]

    # Create new DatFrame with extracted metrics
    df_metrics_full = pd.DataFrame(columns=["_id", *cols])
    for id in id_list:
        df_metrics_full = df_metrics_full.append(pd.DataFrame([[id]], columns=['_id']))
        df_metrics_full.index = df_metrics_full['_id'].tolist()
        for col in metrics:
            try:
                df_metrics_full[f'metrics.{col}'][id] = df_metric[(df_metric["run_id"] == id) & (df_metric["name"] == col)]['values'].values[0]
            except IndexError:
                df_metrics_full[f'metrics.{col}'][id] = float("NaN")
            # Extract float values for test metrics
            try:
                if "test_" in col and len(tmp_vals := df_metrics_full[f'metrics.{col}'][id]) == 1:
                        df_metrics_full[f'metrics.{col}'][id] = tmp_vals[0]
            except TypeError or IndexError:
                pass

    for i in df_m_del_cols:
        if i in df_metrics_full.columns.tolist():
            del df_metrics_full[i]

    df = df_runs.join(df_metrics_full)
    df = df.rename(columns={'_id': 'id'})

    return df.infer_objects()


def _get_dataframe_cached(db: Database,
                          ids: Optional[Union[int, List[int], Tuple[int, ...]]] = None,
                          include_artifacts: bool = False,
                          verify: bool = False,
                          force: bool = False) -> pd.DataFrame:
    """Retrieve DataFrame from MongoDB and cache result locally.

    Caution: When retrieving the full database (using `ids = None`), by default
    it is not verified that the local version is up-to-date with the remote
    version. To verify and possibly re-download the up-to-date database,
    use `verify=True`.

    Args:
        db: Database instance to query.

        ids: Optional run ID or list of IDs to get.
             If None, all IDs are queried.

        include_artifacts: Whether to include a column for artifacts.
                           Artifacts can then be resolved via
                           :func:resolve_artifacts().

        verify: Whether to verify that local version is up-to-date with remote
                database. Only used when `ids=None`.

        force: Whether to force re-loading from MongoDB even if cached
               version is available.

    Returns:
        DataFrame of the corresponding Database.
    """

    filename = f"{db.name}_ids_{ids}_artifacts_{include_artifacts}.pickle"
    get_data = lambda: pickle.dumps(_get_dataframe(db, ids, include_artifacts), protocol=4)
    path = caching.get(filename, get_data, force)

    df_local = pd.read_pickle(path)

    if ids is None and verify and not force:
        print("Verifying local version.")
        ids_local = df_local['id'].tolist()
        ids_remote = _get_run_ids(db)
        if ids_local != ids_remote:
            print("Remote version has been altered. Re-downloading.")
            path = caching.get(filename, get_data, force=True)
            df_local = pd.read_pickle(path)

    return df_local


def get_dataframe(db: Database,
                  ids: Optional[Union[int, List[int], Tuple[int, ...]]] = None,
                  include_artifacts: bool = False,
                  cache: bool = False,
                  **kwargs):
    """Get all experiments from a Sacred Database in a DataFrame.
    Metrics and configurations are concatenated into a single DataFrame for
    convenience.

    Args:
        db: Database instance to query.

        ids: Optional run ID or list of IDs to get. If None, all IDs are queried.

        include_artifacts: Whether to include a column for artifacts.
                           Artifacts can then be resolved via
                           :func:resolve_artifacts().

        cache: Whether to cache the downloaded DataFrame. This can be useful
               when retrieving very large databases. See :func:`_get_dataframe_cached()`
               for details and keyword arguments.

        **kwargs: See :func:`_get_dataframe_cached()`.

    Returns:
        DataFrame of the corresponding Database.
    """
    if cache:
        return _get_dataframe_cached(db, ids, include_artifacts, **kwargs)
    else:
        return _get_dataframe(db, ids, include_artifacts)


def get_run_names(db: Database):
    """Get a DataFrame containing only the run ID experiment names.

    Args:
        db:

    Returns:
        DataFrame with run IDs and experiment names.
    """
    df_runs = _get_dataframe_query(db, {}, "runs")
    df_names = df_runs["experiment"].map(lambda x: x['name'])
    df_names.index = df_runs['_id'].tolist()
    return df_names


def resolve_artifacts(db: Database, df: pd.DataFrame) -> pd.DataFrame:
    """Maps the artifact column as retrieved from the database in
    get_df_full(..., return_artifacts=True) into a dictionary of the form
    {filename1: filecontent1, filename2:  filecontent2....}.

    Usage:

        df = mdbh.get_df_full(db, return_artifacts=True)
        df = mdbh.resolve_artifacts(db, df)

    Args:
        db: Database instance.
        df: The Dataframe with 'artifacts' column.

    Returns:
        The Dataframe with a transformed column
    """
    def map_artifacts(db: Database, x):
        return {artifact["name"]: _get_gridfs_data(db, artifact["file_id"]) for artifact in x}
    df["artifacts"] = df["artifacts"].map(lambda x: map_artifacts(db, x))
    return df


def get_artifact_names(db: Database, id: int) -> List[str]:
    """Get the names of the artifacts from a run ID.

    Args:
        db: Database instance.
        id: Run ID.

    Returns:
        List of names of the ID's artifacts.
    """
    df = get_dataframe(db, [id], include_artifacts=True)
    return [a['name'] for a in df.loc[id]['artifacts']]


def get_artifact(db: Database,
                 id: int,
                 name: str,
                 force: bool = False) -> Path:
    """Get path to artifact from a Database and run ID.
    The artifacts are downloaded and cached by default.

    Args:
        db: Database instance to query.

        id: Run ID to get artifact from.

        name: Name of artifact to get.
              To obtain a list of available names, use
              :func:`get_artifact_names()`

        force: Whether to force downloading of the artifact, even if cached
               version is available.

    Returns:
        Path to the downloaded or cached artifact.
    """
    def _get_data():
        """Download artifact from MongoDB."""
        df = get_dataframe(db, [id], include_artifacts=True)
        artifacts = [a for a in df.loc[id]['artifacts'] if a['name'] == name]
        if len(artifacts) == 0:
            raise ValueError(f"No artifact found with name {name} for run ID {id}.")
        if len(artifacts) > 1:
            raise ValueError(f"Found more than one artifact with name {name} for run ID {id}.")
        artifact = artifacts[0]
        print(f"Downloading artifact from MongoDB.")
        return _get_gridfs_data(db, artifact['file_id'])

    filename = f"{db.name}_{id}_{name}"
    return caching.get(filename, _get_data, force)


def add_artifact(db: Database,
                 id: int,
                 path: Union[str, Path],
                 name: Optional[str] = None):
    """Adds an artifact to an existing Sacred experiment with run ID.

    Args:
        db: Database instance to query.

        id: Run ID to add artifact to.

        path: System path to the file to add as an artifact.

        name: Optionally specify name of artifact.
              If None, name is used from specified `path`.
    """
    if id not in _get_run_ids(db):
        raise ValueError(f"ID {id} not found in database.")

    path = Path(path)
    name = name or path.name

    # Get artifacts already part of the experiment
    query = {"_id": {"$in": [id]}}
    res = db["runs"].find(query)
    artifacts = res[0]['artifacts']

    # Check if artifact name is available
    if name in [a['name'] for a in artifacts]:
        raise ValueError(f"Name {name} already used by another artifact.")

    # Add artifact to GridFS, use layout specified by Sacred
    db_filename = f"artifact://runs/{id}/{name}"
    object_id = data._put_gridfs_data(db,
                                      path,
                                      filename=db_filename,
                                      metadata=None,
                                      content_type=None)

    # Append to existing artifacts and update database entry
    artifacts.append(dict(name=name, file_id=object_id))
    db["runs"].update_one(query, {"$set": {"artifacts": artifacts}})

    return


def get_metric(db: Database, id: int, metric: str) -> dict:
    """Get metric from a Sacred experiment ID.

    Args:
        db: PyMongo Database instance.

        id: Run ID of experiment.

        metric: Name of the metric to get.

    Returns:
        dict, metric dictionary containing name, value, steps, etc.
    """
    if type(db) != Database:
        raise TypeError("db needs to be a PyMongo Database instance. "
                        "Use mdbh.get_mongodb")

    metrics = db.get_collection('metrics')
    return metrics.find_one({"run_id": id, "name": metric})



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import setuptools
from setuptools import setup
import subprocess

# Get Version from Pipeline, ignore leading 'v'
if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG'][1:]
elif os.environ.get('CI_JOB_ID'):
    version = os.environ['CI_JOB_ID']

# For local builds:
else:
    try:
        # Get latest git tag
        result = subprocess.run("git describe --tags", shell=True, stdout=subprocess.PIPE)
        version = result.stdout.decode('utf-8')[1:-1] + "-local"
    except:
        version = "local"

# Write version file for module to import
with open("mdbh/_version.py", "w") as f:
    f.write(f"__version__='{version}'")

# Load README as full description
short_description = "A MongoDB, Sacred, and Omniboard helper."
try:
    with open("README.md", "r") as f:
        long_description = f.read()
except:
    long_description = short_description

setup(
    python_requires='>=3.8',
    name='mdbh',
    version=version,
    description=short_description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/MaxSchambach/mdbh',
    author='Maximilian Schambach',
    author_email='schambach@kit.edu',
    install_requires=[
        'pymongo >= 3.0',
        'pandas',
        'stopit >= 1.1.2',
    ],
    extras_require={
        "plot":  ["numpy", "matplotlib"],
    },
    license='GNU General Public License v3 (GPLv3)',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    zip_safe=True)



# -*- coding: utf-8 -*-1.3
# Copyright (C) 2020  The MDBH Authors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path

from datetime import datetime

from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import gridfs
from bson.objectid import ObjectId

from pymongo.database import Collection
from pymongo.database import Database
from pymongo.results import DeleteResult

from mdbh import caching
from mdbh.environ import get_data_collection


def _delete_gridfs_data(db: Database,
                        id: Union[str, ObjectId]):
    """Delete GridFS object.

    Args:
        db: MongoDB Database instance.

        id: ObjectID of the GridFS data entry.
    """
    id = id if isinstance(id, ObjectId) else ObjectId(id)
    fs = gridfs.GridFS(db)
    fs.delete(id)
    return


def _get_gridfs_data(db: Database,
                     id: Union[str, ObjectId]) -> bytes:
    """Retrieves file content from GridFS by ObjectID.
    E.g., this corresponds to the way Sacred saves artifacts.

    Args:
        db: Database instance to query.

        id: Object id of the GridFS data entry.

    Returns:
        The binary file content.
    """
    id = id if isinstance(id, ObjectId) else ObjectId(id)
    gr = gridfs.GridFS(db)
    return gr.get(id).read()


def _put_gridfs_data(db: Database,
                     path: Union[str, Path],
                     **kwargs) -> ObjectId:
    """Put binary data into GridFS.

    Args:
        db: MongoDB Database instance.

        path: System path to the file

        **kwargs: Passed to GridFS.put()

    Returns:
        ObjectID of the created GridFS entry.
    """
    with open(path, 'br') as fh:
        # Add file to GridFS
        fs = gridfs.GridFS(db)
        new_data_id = fs.put(fh, **kwargs)

    return new_data_id


def _get_data_collection_entry(db: Database,
                               id: Union[str, ObjectId]) -> Tuple[Collection, dict]:
    """Get entry and collection from binary data collection.

    Args:
        db: MongoDB Database instance.

        id: ObjectID of the entry.

    Returns:
        collection, entry
    """
    id = id if isinstance(id, ObjectId) else ObjectId(id)
    collection = db[get_data_collection()]
    entry = collection.find_one({"_id": id})
    return collection, entry


def delete_data(db: Database,
                id: Union[str, ObjectId]) -> DeleteResult:
    """Delete data and object entry.
    To delete all data, use
    `[delete_data(db, id) for id in [d['_id'] for d in list_data(db)]]`

    Args:
        db: MongoDB Database instance.
        id: ObjectId of the object to delete.

    Returns:
        DeleteResult
    """
    id = id if isinstance(id, ObjectId) else ObjectId(id)
    collection, entry = _get_data_collection_entry(db, id)

    if entry is not None:
        # Delete corresponding binary data
        data_id = entry["data_id"]
        _delete_gridfs_data(db, data_id)
    else:
        print(f"No object with matching ID found.")

    # Delete entry
    return collection.delete_one({"_id": id})


def find_data(db: Database,
              query: Optional[dict] = None) -> List[dict]:
    """Find data in binary data collection.
    E.g., to query a specific object id, use `query = {'id'=ObjectId(...)}`.

    Args:
        db: MongoDB Database instance.

        query: MongoDB query.

    Returns:
        List of retrieved documents matching the query.
    """

    query = query or {}
    if "id" in query.keys():
        query['_id'] = query.pop('id')

    return list(db[get_data_collection()].find(query))


def get_data(db: Database,
             id: Union[str, ObjectId],
             force: bool = False) -> Path:
    """Get path to artifact from a Database and object ID.
    The artifacts are downloaded and cached by default.

    Args:
        db: Database instance to query.

        id: Object ID to get binary data from.

        force: Whether to force downloading of the artifact, even if cached
               version is available.

    Returns:
        Path to the downloaded or cached artifact.
    """
    id = id if isinstance(id, ObjectId) else ObjectId(id)
    collection, entry = _get_data_collection_entry(db, id)
    filename = f"{db.name}_{get_data_collection()}_id_{id}_{entry['filename']}"
    data_id = entry['data_id']
    _get_data = lambda: _get_gridfs_data(db, data_id)

    return caching.get(filename, _get_data, force)


def list_data(db: Database) -> List[dict]:
    """List all binary data in data collection.

    Args:
        db: MongoDB Database instance.

    Returns:
        List of retrieved documents as dict.
    """
    return find_data(db, {})


def put_data(db: Database,
             path: Union[str, Path],
             name: Optional[str] = None,
             **kwargs) -> ObjectId:
    """Put binary data into a database.
    The collection name is determined by the MDBH_DATA_COLLECTION environment
    variable, see `mdbh.environ`.

    See Also:
        To read the data from the database with caching, see :func:`get_data`.
        To delete the data, see :func:`delete_data`.
        To find data, see :func:`find_data`.

    Args:
        db: MongoDB Database instance

        path: System path to file containing the data.

        name: Optional name for database entry.

        **kwargs: Optional metadata arguments to identify the file.

    Returns:
        ObjectId instance of the new database entry.
    """
    if isinstance(path, str):
        path = Path(path)

    # Upload data to GridFS database
    data_id = _put_gridfs_data(db, path)

    # Insert into collection
    res = db[get_data_collection()].insert_one(
        dict(filename=path.name, name=name, date=datetime.now().isoformat(),
             data_id=data_id, **kwargs))

    return res.inserted_id


def replace_data(db: Database,
                 id: Union[str, ObjectId],
                 path: Union[str, Path]):
    """Replace binary data associatet with entry id.

    Args:
        db: MongoDB Database instance.

        id: ObjectId of document.

        path: Path to new data file.
    """
    id = id if isinstance(id, ObjectId) else ObjectId(id)
    collection, entry = _get_data_collection_entry(db, id)

    if entry is not None:
        # Delete corresponding binary data
        data_id = entry["data_id"]
        _delete_gridfs_data(db, data_id)
    else:
        print(f"No object with matching ID found.")

    if isinstance(path, str):
        path = Path(path)

    # Upload new data to GridFS
    new_data_id = _put_gridfs_data(db, path)

    # Update/replace document
    entry['data_id'] = new_data_id
    entry['filename'] = path.name
    entry['date_modified'] = datetime.now().isoformat()
    collection.replace_one({"_id": id}, entry)
    return
